/* Mixmaster version 3  --  (C) 1999 Anonymizer Inc.

   Mixmaster may be redistributed and modified under certain conditions.
   This software is distributed on an "AS IS" basis, WITHOUT WARRANTY OF
   ANY KIND, either express or implied. See the file COPYRIGHT for
   details.

   Prepare messages for remailer chain
   $Id: chain.c,v 1.22 1999/06/04 00:33:09 um Exp $ */


#include "mix3.h"
#include <string.h>
#include <ctype.h>
#include <assert.h>

void clienterr(BUFFER *msgbuf, char *err)
{
  if (msgbuf) {
    buf_sets(msgbuf, "Error: ");
    buf_appends(msgbuf, err);
  } else
    errlog(ERRORMSG, "%s\n", err);
}

int chain_select(int hop[], char *chainstr, int maxrem, REMAILER *remailer,
		 int type, BUFFER *feedback)
{
  int len = 0;
  int i, j, k;
  BUFFER *chain, *selected, *addr;
  chain = buf_new();
  selected = buf_new();
  addr = buf_new();

  if (chainstr == NULL || chainstr[0] == '\0')
    buf_sets(chain, CHAIN);
  else
    buf_sets(chain, chainstr);

  /* put the chain backwards: final hop is in hop[0] */

  for (i = chain->length; i >= 0; i--)
    if (i == 0 || chain->data[i - 1] <= ' ' || chain->data[i - 1] == ','
	|| chain->data[i - 1] == ';' || chain->data[i - 1] == ':') {
      for (j = i; isspace(chain->data[j]);)	/* ignore whitespace */
	j++;
      if (chain->data[j] == '\0')
	break;

      if (chain->data[j] == '*')
	k = 0;
#if 0
      else if (isdigit(chain->data[j]))
	k = atoi(chain->data + j);
#endif
      else {
	buf_sets(selected, chain->data + j);
	rfc822_addr(selected, addr);
	buf_clear(selected);
	buf_getline(addr, selected);
	if (!selected->length)
	  buf_sets(selected, chain->data + j);

	for (k = 0; k < maxrem; k++)
	  if (((remailer[k].flags.mix && type == 0) ||
	       (remailer[k].flags.cpunk && type == 1) ||
	       (remailer[k].flags.newnym && type == 2)) &&
	      (streq(remailer[k].name, selected->data) ||
	       strieq(remailer[k].addr, selected->data) ||
	       (selected->data[0] == '@' && strifind(remailer[k].addr,
					    selected->data))))
	    break;
      }
      if (k < 0 || k >= maxrem) {
	buf_appendf(feedback, "No such remailer: %b", selected);
	buf_nl(feedback);
#if 0
	k = 0;
#else
	len = -1;
	goto end;
#endif
      }
      hop[len++] = k;
      if (len > 20) {
	buf_appends(feedback, "Chain too long.\n");
	break;
      }
      if (i > 0)
	chain->data[i - 1] = '\0';
    }
end:
  buf_free(chain);
  buf_free(selected);
  buf_free(addr);
  return len;
}

int chain_randfinal(int type, REMAILER *remailer, int maxrem, int rtype)
{
  int num = 0;
  int i;
  int t;

  t = rtype;
  if (rtype == 2)
    t = 1;

  /* select a random final hop */
  for (i = 1; i < maxrem; i++) {
    if (((remailer[i].flags.mix && rtype == 0) ||
	 (remailer[i].flags.ek && rtype == 1) ||
	 (remailer[i].flags.newnym && rtype == 2)) &&
	remailer[i].info[t].reliability >= 100 * RELFINAL &&
	remailer[i].info[t].latency <= MAXLAT &&
	(type == MSG_NULL || !remailer[i].flags.middle) &&
	(remailer[i].flags.post || type != MSG_POST))
      num++;
  }
  if (num == 0)
    i = -1;
  else {
    do
      i = rnd_number(maxrem) + 1;
    while (!(((remailer[i].flags.mix && rtype == 0) ||
	      (remailer[i].flags.ek && rtype == 1) ||
	      (remailer[i].flags.newnym && rtype == 2)) &&
	     remailer[i].info[t].reliability >= 100 * RELFINAL &&
	     remailer[i].info[t].latency <= MAXLAT &&
	     (type == MSG_NULL || !remailer[i].flags.middle) &&
	     (remailer[i].flags.post || type != MSG_POST)));
  }
  return (i);
}

int chain_rand(REMAILER *remailer, int maxrem,
	       int thischain[], int chainlen, int t)
     /* set random chain. returns 0 if not random, 1 if random, -1 on error */
{
  int hop;
  int err = 0;

  assert(t == 0 || t == 1);

  for (hop = 0; hop < chainlen; hop++)
    if (thischain[hop] == 0) {
      int select[MAXREM];
      int randavail = 0;
      int i;

      err = 1;
      for (i = 1; i < maxrem; i++)
	select[i] = ((remailer[i].flags.mix && t == 0) ||
		     (remailer[i].flags.ek && t == 1))
	  && remailer[i].info[t].reliability >= 100 * MINREL &&
	  remailer[i].info[t].latency <= MAXLAT,
	  randavail += select[i];

      for (i = hop - DISTANCE; i <= hop + DISTANCE; i++)
	if (i >= 0 && i < chainlen && select[thischain[i]])
	  select[thischain[i]] = 0, randavail--;

      if (randavail < 1) {
	err = -1;
	goto end;
      }
      do
	thischain[hop] = rnd_number(maxrem - 1) + 1;
      while (!select[thischain[hop]]);
    }
end:
  return (err);
}

int mix_encrypt(int type, BUFFER *message, char *chainstr, int numcopies,
		BUFFER *chainlist)
{
  return (mix2_encrypt(type, message, chainstr, numcopies, chainlist));
}
